
/**************************************************************************
c2tex - Konverze C++ do TeXu
Jiri Bittner, Cerven 1996
**************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>


#define	KEYWORD1	1	// kody pro jednotlive znacky
#define	KEYWORD2	2
#define	SYMBOL		3
#define	COMMENT        4
#define	NORMAL		5

#define	WORD	10

#define	S_LEN	256

int usePascal=0;

char comment_1b='/',comment_2b='*',comment_1e='*',comment_2e='/';

class BaseC
{

};


typedef double Real;


class OptionsC : public BaseC
{
public:

  int    number;
  char **strings;
  
//  OptionsC() { number=0; }
  OptionsC(int n,char **a) { init(n,a); }

  void init(int n,char **s) { number=n; strings=s; }
  int isOption(char *string);
  int isOption(char *string,int *number);
  int getParam(char *s,int n=1);

  int getOptionValue(char *string,int *result);
  int getOptionValue(char *string,Real *r);
  int getOptionValue(char *string,char *result);

};


class LexanC: public BaseC
{
public:
FILE *file;

int 	nsep,nkey[2];
char	*separators;
char	**key[2];

LexanC(){init("default.lex");}

int	isSeparator(int c){ return strchr(separators,c)!=NULL; }
int	isKey(char *s);

int	init(const char *lexname);
int 	open(const char *filename);

int	readWord(FILE *f,char *string);

int	read(char *string);
void 	close(){fclose(file);}

virtual	ostream& operator>>(ostream &s) { 
	s<<nsep<<"\n"<<nkey[0]<<"\n"<<nkey[1]<<"\n";
	return s;
					}
~LexanC(){fclose(file);}
};

class TexC: public BaseC
{
public:
  FILE *file;
  int numbers;
  int lnumber;
  
  TexC(){numbers=0; init("output.tex");}
  TexC(int n, const char *s){
    lnumber=0;
    numbers=n; init(s);}
  
  int init(const char *filename);
  int write(const int code,const char *string);
  ~TexC() { /* fprintf(file,"\n\\bye\n"); */ fclose(file);}

};








int OptionsC::isOption(char *s)
{
int found=0;
 
 for (int i=1;i<number;i++)
   if (strcmp(s,strings[i])==0) { found = i; break;}

 return found;
}

int OptionsC::isOption(char *s,int *n)
{
int found=0;

for (int i=1;i<number;i++)
  if (strcmp(s,strings[i])==0)
    {
      if (i+1<number)
	{
	  *n=atoi(strings[i+1]);
	  found = i;
	  break;
	}
    }

return found;
}


// search for occurence of non-option
int OptionsC::getParam(char *s,int n)
{
int i,k=1,found=0;

for (i=1;i<number;i++)
  if (strings[i][0]!='-')
    if (k==n) { found=1; break;}
       else
	 k++;

if (found) strcpy(s,strings[i]);

return found;
}



int OptionsC::getOptionValue(char *s,int *result)
{
int found=0;
char *pp;

for (int i=1;i<number;i++)
  {
    if (strstr(strings[i],s))
    {
      pp=strings[i]+strlen(s);
      *result=atoi(pp);
      found = i;
      break;
    }
  }
return found;
}

int
OptionsC::getOptionValue(char *s,Real *result)
{
int found=0;
char *pp;

for (int i=1;i<number;i++)
  {
    if (strstr(strings[i],s))
    {
      pp=strings[i]+strlen(s);
      *result=atof(pp);
      found = i;
      break;
    }
  }
return found;
}

int OptionsC::getOptionValue(char *s,char *result)
{
int found=0;
char *pp;

for (int i=1;i<number;i++)
  {
    if (strstr(strings[i],s))
    {
      pp=strings[i]+strlen(s);
      strcpy(result,pp);
      found = i;
      break;
    }
  }
return found;
}


int TexC::write(const int code,const char *string)
{
static int line=1;
static int i=0;
const char *sp;

switch(code) {
  
 case KEYWORD1 : fprintf(file,"\\keya{}"); break;
 case KEYWORD2 : fprintf(file,"\\keyb{}"); break;
 case COMMENT  : fprintf(file,"\\comment{}"); break;
 case SYMBOL   : fprintf(file,"\\symbol{}"); break;
 case NORMAL   : fprintf(file,"\\normal{}"); break;
}
 
 for(sp=string;*sp!=0;sp++,i++)
   {
     if (line) {
       fprintf(file,"\\leftline{"); line=0;
       if (strstr(string, "Algorithm") == string)
	 lnumber = 0;
       if (numbers && lnumber)
	 fprintf(file,"%2d:\\ \\ \\ ",lnumber);
       lnumber++;
     }
     
     switch(*sp) {
       
#define TAB_SIZE	8
     case '\t' :	
       //		fprintf(file,"\\tab"); break;
       int n,j;
       n=((i/TAB_SIZE)+1)*TAB_SIZE-i;
       for(j=0;j<n;j++) fprintf(file,"\\ ");
       break;
       
       
     case '\n': 	fprintf(file," }\n");
       line=1; i=0;
       break;
     case ' ' :	fprintf(file,"\\ ");
      break;
      case '&' :	fprintf(file,"\\&"); break;
      case '#' :	fprintf(file,"\\#"); break;
	//      case '_' :	fprintf(file,"\\_{}"); break;
      case '_' :	fprintf(file,"_"); break;
      case '^' :	fprintf(file,"\\^{}"); break;
      case '%' :	fprintf(file,"\\%%"); break;
      case '{' :	fprintf(file,"$\\{$"); break;
      case '}' :	fprintf(file,"$\\}$"); break;
	//  case '\\' :	fprintf(file,"$\\backslash$"); break;
	//  case '$' :	fprintf(file,"\\$"); break;
    //  case '<' :	fprintf(file,"\\<"); break;
    //  case '>' :	fprintf(file,"\\>"); break;

  case '~' :	fprintf(file,"\\~{}"); break;
      
    
  default  :	fputc(*sp,file);
  }
}
return 1;
}


int	LexanC::isKey(char *s)
{
int i,j;

for (i=0;i<2;i++)
 for(j=0;j<nkey[i];j++)
  if(strcmp(s,key[i][j])==0) return i+1;

return 0;
}


int 	LexanC::readWord(FILE *f,char *s)
{
int	i=0;
int	c;

do
{
	c=fgetc(f);
	if (c==EOF) return 0;
	if (!isSeparator(c)) break;
} while(1);


do
{
	s[i++]=c;
	c=fgetc(f);
	if (c==EOF) break;
	if (isSeparator(c)) break;
} while(1);

s[i]=0;
return 1;
}


int	LexanC::init(const char *lexname)
{
FILE *f;

if ((f=fopen(lexname,"rt"))==NULL) return 0;


#define 	NKEY	512

separators = new char[256];
key[0]   = new char*[NKEY];
key[1]   = new char*[NKEY];

nsep=0;
separators[nsep++]=' ';
separators[nsep++]='\n';
separators[nsep++]='\t';

fgets(separators+nsep,100,f);
nsep=strlen(separators);

int i=0,type=0;
char s[64];

while(readWord(f,s))
	{
	if (strcmp(s,"KEYWORDS2")==0) 
	{
	    printf("KEYWORDS2\n");
	    nkey[type]=i; 
	    type++; 
	    i=0;  
	}
	else
	{
		printf("%s\n",s);	
		key[type][i]=new char[strlen(s)+1];
		strcpy(key[type][i],s);
		i++;
	}
	}

nkey[type]=i; 

fclose(f);

 
return 1;
}

int 	LexanC::open(const char *filename)
{
file=fopen(filename,"rt");
return file!=NULL;
}

int	LexanC::read(char *s)
{
int	i=0;
int	type=EOF;
int	wasSlash=0;
static	int	c=-2;

if (c==-2)  c=fgetc(file);

if (c==EOF) return c;

do{
if (isSeparator(c)) {
	type=SYMBOL;
	s[i++]=c;

	if(wasSlash) {
	  if(c==comment_1b)  // line comment
	    { 
	      if (i>2) { c=s[i-2]; fseek(file,-1,SEEK_CUR); s[i-2]=0; 
	      return type; }
	      fgets(s+i,S_LEN,file); 
	      c=fgetc(file);
	      return COMMENT;
	    }
	  else
	    if(c==comment_2b)  // line comment
	      { 
		if (i>2) { c=s[i-2]; fseek(file,-1,SEEK_CUR); s[i-2]=0; 
		return type; }
		c=fgetc(file);
		if(c!=EOF)
		  {
		    s[i++]=c;
		    int count=1;
			do{
			  c=fgetc(file);
			  if(c==EOF) break;
			  s[i++]=c;
			  if (s[i-2]==comment_1e && s[i-1]==comment_2e) count--;
			  if (s[i-2]==comment_1b &&s[i-1]==comment_2b) count++;
			} while(count!=0);
		      }
		    s[i]=0;
		    c=fgetc(file);
		    return COMMENT;
		  }
	    }
	  else
	    wasSlash=(c==comment_1b);	
} else
  if (c=='`') {
    if (i>=1) {
      s[i]=0;
      return type;
    }
    c=fgetc(file);
    while(c!='`') {
      if (c==EOF) break;
      s[i++]=c;
      c=fgetc(file);
    }
    s[i]=0;
    c=fgetc(file);
    return WORD;
  } else {
    if (type==EOF) type=WORD;
    break;
  }
} while((c=fgetc(file))!=EOF);
 
if (c!=EOF)
if (type==WORD)
{
do
{
s[i++]=c;
c=fgetc(file);
if (c==EOF) break;
} while(!isSeparator(c));

s[i]=0;
int k=isKey(s);
if (k==0) return NORMAL;
return k;
}

s[i]=0;
return type;

}

int TexC::init(const char *filename)
{
file=fopen(filename,"wt");
fprintf(file,"\\input default.mac\n");
return file!=NULL;
}


void Help()
{
  printf("Syntax : c2tex in_file [out_file] [-usePascal] [-numbers]\n\n");
  printf("Default out_file is output.tex.\n");
  printf("-usePascal     use usePascal comments (* *).\n");
  printf("-numbers    print line numbers.\n");
  exit(1);
}

int main(int argc, char **argv)
{
char	s[16384]; // docasny buffer

OptionsC options(argc,argv);
 
 if (argc<2) Help();
 if (argc>=3 && (*argv[2]!='-') )
   strcpy(s,argv[2]);
 else
   strcpy(s,"output.tex");
 
 if (options.isOption("-usePascal")) {
    usePascal=1;
    comment_1b='(';
    comment_2b='*';
    comment_1e='*';
    comment_2e=')';
 }

 int numbers;
 
 numbers = options.isOption("-numbers");
 
 
  LexanC	lexan;
  TexC	tex(numbers, s);
  lexan.open(argv[1]);

  int	t;
  cout<<tex.numbers;

  
  while((t=lexan.read(s))!=EOF)
    {
      //printf("Type : %d - %s\n",t,s);
      printf("%s",s);
      tex.write(t,s);
      //getchar();
    }

return 0;
}





