Algorithm: Traversal of the kD-tree applying batching of multiple queries
TraversalStack.Push(kDTree.Root);
while ( not TraversalStack.Empty() or
        not QueryQueue.Empty() ) {
  //---- PART 1: processing finished occlusion queries
  while ( not QueryQueue.Empty() and
	  (ResultAvailable(QueryQueue.Front()) or
	   TraversalStack.Empty()) ) {
    N = QueryQueue.Dequeue();
    // wait if result not available
    visiblePixels = GetOcclussionQueryResult(N);
    if ( visiblePixels > VisibilityThreshold ) {
      PullUpVisibility(N);
      TraverseNode(N);
    }
  }
  //---- PART 2: kd-tree traversal
  if ( not TraversalStack.Empty() ) {
    N = TraversalStack.Pop();
    if ( InsideViewFrustum(N) ) {
      // identify previously visible nodes
      wasVisible = N.visible && (N.lastVisited == frameID -1);
      // identify previously opened nodes
      opened = wasVisible && !IsLeaf(N);
      // reset node's visibility classification
      N.visible = false;
      // update node's visited flag
      N.lastVisited = frameID;
      // skip testing all previously opened nodes
      if ( !opened ) {
		  // add to the pending queue instead of immediate query
		  if (PendingQueue.size() < n)
			  PendingQueue.Enqueue(N);
		  else
		  {
			  IssueMultipleOcclusionQueries(PendingQueue);
			  AddToQueryQueue(PendingQueue, QueryQueue);
			  PendingQueue.clear();
		  }
      }
      // traverse a node unless it was invisible
      if ( wasVisible )
	TraverseNode(N);
    }
  }

  // add rest to query queue
  IssueMultipleOcclusionQueries(PendingQueue);
  AddToQueryQueue(PendingQueue, QueryQueue);
  PendingQueue.clear();
}
TraverseNode(N) {
  if ( IsLeaf(N) )
    Render(N);
  else
    TraversalStack.PushChildren(N);
}
PullUpVisibility(N) {
  while (!N.visible) { N.visible = true; N = N.parent; }
}
